@@ -19,7 +19,7 @@ enum {
   TITLE_COLUMN,
   FILENAME_COLUMN,
   FUNC_COLUMN,
-  ITALIC_COLUMN,
+  STYLE_COLUMN,
   NUM_COLUMNS
 };
 
@@ -94,15 +94,15 @@ window_closed_cb (GtkWidget *window, gpointer data)
 {
   CallbackData *cbdata = data;
   GtkTreeIter iter;
-  gboolean italic;
+  PangoStyle style;
 
   gtk_tree_model_get_iter (cbdata->model, &iter, cbdata->path);
   gtk_tree_model_get (GTK_TREE_MODEL (cbdata->model), &iter,
-		      ITALIC_COLUMN, &italic,
+		      STYLE_COLUMN, &style,
 		      -1);
-  if (italic)
+  if (style == PANGO_STYLE_ITALIC)
     gtk_tree_store_set (GTK_TREE_STORE (cbdata->model), &iter,
-			ITALIC_COLUMN, !italic,
+			STYLE_COLUMN, PANGO_STYLE_NORMAL,
 			-1);
 
   gtk_tree_path_free (cbdata->path);
@@ -586,7 +586,7 @@ row_activated_cb (GtkTreeView       *tree_view,
 		  GtkTreeViewColumn *column)
 {
   GtkTreeIter iter;
-  gboolean italic;
+  PangoStyle style;
   GDoDemoFunc func;
   GtkWidget *window;
   GtkTreeModel *model;
@@ -597,14 +597,14 @@ row_activated_cb (GtkTreeView       *tree_view,
   gtk_tree_model_get (GTK_TREE_MODEL (model),
 		      &iter,
 		      FUNC_COLUMN, &func,
-		      ITALIC_COLUMN, &italic,
+		      STYLE_COLUMN, &style,
 		      -1);
 
   if (func)
     {
       gtk_tree_store_set (GTK_TREE_STORE (model),
 			  &iter,
-			  ITALIC_COLUMN, !italic,
+			  STYLE_COLUMN, (style == PANGO_STYLE_ITALIC ? PANGO_STYLE_NORMAL : PANGO_STYLE_ITALIC),
 			  -1);
       window = (func) (gtk_widget_get_toplevel (GTK_WIDGET (tree_view)));
       
@@ -700,7 +700,7 @@ create_tree (void)
 
   Demo *d = testgtk_demos;
 
-  model = gtk_tree_store_new (NUM_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER, G_TYPE_BOOLEAN);
+  model = gtk_tree_store_new (NUM_COLUMNS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER, G_TYPE_INT);
   tree_view = gtk_tree_view_new ();
   gtk_tree_view_set_model (GTK_TREE_VIEW (tree_view), GTK_TREE_MODEL (model));
   selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree_view));
@@ -723,7 +723,7 @@ create_tree (void)
 			  TITLE_COLUMN, d->title,
 			  FILENAME_COLUMN, d->filename,
 			  FUNC_COLUMN, d->func,
-			  ITALIC_COLUMN, FALSE,
+			  STYLE_COLUMN, PANGO_STYLE_NORMAL,
 			  -1);
 
       d++;
@@ -742,7 +742,7 @@ create_tree (void)
 			      TITLE_COLUMN, children->title,
 			      FILENAME_COLUMN, children->filename,
 			      FUNC_COLUMN, children->func,
-			      ITALIC_COLUMN, FALSE,
+			      STYLE_COLUMN, PANGO_STYLE_NORMAL,
 			      -1);
 	  
 	  children++;
@@ -751,14 +751,10 @@ create_tree (void)
 
   cell = gtk_cell_renderer_text_new ();
 
-  g_object_set (cell,
-                "style", PANGO_STYLE_ITALIC,
-                NULL);
-  
   column = gtk_tree_view_column_new_with_attributes ("Widget (double click for demo)",
 						     cell,
 						     "text", TITLE_COLUMN,
-						     "style_set", ITALIC_COLUMN,
+						     "style", STYLE_COLUMN,
 						     NULL);
   
   gtk_tree_view_append_column (GTK_TREE_VIEW (tree_view),
@@ -786,6 +782,8 @@ create_tree (void)
 
   gtk_widget_grab_focus (tree_view);
 
+   g_object_unref (model);
+
   return box;
 }
 
@@ -869,7 +867,7 @@ main (int argc, char **argv)
   
   window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title (GTK_WINDOW (window), "GTK+ Code Demos");
-  g_signal_connect (window, "destroy",
+  g_signal_connect_after (window, "destroy",
 		    G_CALLBACK (gtk_main_quit), NULL);
 
   hbox = gtk_hbox_new (FALSE, 0);
@@ -885,13 +883,15 @@ main (int argc, char **argv)
 			    create_text (&info_buffer, FALSE),
 			    gtk_label_new_with_mnemonic ("_Info"));
 
+  tag = gtk_text_buffer_create_tag (info_buffer, "title",
+                                    "font", "Sans 18",
+                                    NULL);
+   g_object_unref (info_buffer);
+
   gtk_notebook_append_page (GTK_NOTEBOOK (notebook),
 			    create_text (&source_buffer, TRUE),
 			    gtk_label_new_with_mnemonic ("_Source"));
 
-  tag = gtk_text_buffer_create_tag (info_buffer, "title",
-                                    "font", "Sans 18",
-                                    NULL);
 
   tag = gtk_text_buffer_create_tag (source_buffer, "comment",
 				    "foreground", "DodgerBlue",
@@ -914,6 +914,7 @@ main (int argc, char **argv)
 				    "weight", PANGO_WEIGHT_BOLD,
  				    "foreground", "DarkGoldenrod4",
                                     NULL);
+   g_object_unref (source_buffer);
   
   gtk_window_set_default_size (GTK_WINDOW (window), 600, 400);
   gtk_widget_show_all (window);
