@@ -29,14 +29,32 @@
 #include <config.h>
 #include <stdlib.h>
 
+#ifndef _MSC_VER
 #define _WIN32_WINNT 0x0500
 #define WINVER _WIN32_WINNT
+#endif
 
 #include "gdk.h"
 #include "gdkprivate-win32.h"
 #include "gdkinput-win32.h"
 
 #if defined(_MSC_VER) && (WINVER < 0x0500)
+
+typedef struct
+{
+  UINT cbSize;
+  HWND hwnd;
+  DWORD dwFlags;
+  UINT uCount;
+  DWORD dwTimeout;
+} FLASHWINFO;
+
+#define FLASHW_STOP 0
+#define FLASHW_CAPTION 1
+#define FLASHW_TRAY 2
+#define FLASHW_ALL (FLASHW_CAPTION|FLASHW_TRAY)
+#define FLASHW_TIMER 4
+
 #define GetAncestor(hwnd,what) _gdk_win32_get_ancestor_parent (hwnd)
 
 static HWND
@@ -83,6 +101,10 @@ static void gdk_window_impl_win32_finalize   (GObject                 *object);
 
 static gpointer parent_class = NULL;
 
+static void     update_style_bits         (GdkWindow *window);
+static gboolean _gdk_window_get_functions (GdkWindow     *window,
+                                           GdkWMFunction *functions);
+
 #define WINDOW_IS_TOPLEVEL(window)		   \
   (GDK_WINDOW_TYPE (window) != GDK_WINDOW_CHILD && \
    GDK_WINDOW_TYPE (window) != GDK_WINDOW_FOREIGN)
@@ -131,6 +153,7 @@ gdk_window_impl_win32_init (GdkWindowImplWin32 *impl)
   impl->hicon_big = NULL;
   impl->hicon_small = NULL;
   impl->hint_flags = 0;
+  impl->type_hint = GDK_WINDOW_TYPE_HINT_NORMAL;
   impl->extension_events_selected = FALSE;
 }
 
@@ -263,7 +286,6 @@ gdk_window_impl_win32_get_visible_region (GdkDrawable *drawable)
 {
   GdkWindowImplWin32 *impl = GDK_WINDOW_IMPL_WIN32 (drawable);
   GdkRectangle result_rect;
-  HDC hdc;
 
   result_rect.x = 0;
   result_rect.y = 0;
@@ -272,25 +294,6 @@ gdk_window_impl_win32_get_visible_region (GdkDrawable *drawable)
 
   gdk_rectangle_intersect (&result_rect, &impl->position_info.clip_rect, &result_rect);
 
-  /* take this win32 specific part into account (smaller when obscured) */
-  hdc = GetDC (GDK_DRAWABLE_IMPL_WIN32_HANDLE (impl));
-  if (hdc)
-    {
-      RECT r;
-      if (SIMPLEREGION == GetClipBox (hdc, &r))
-        {
-          GdkRectangle gr;
-
-          gr.x = r.left + impl->position_info.x_offset;
-          gr.y = r.top + impl->position_info.y_offset;
-          gr.width = r.right - r.left;
-          gr.height = r.bottom - r.top;
-
-          gdk_rectangle_intersect (&result_rect, &gr, &result_rect);
-        }
-      ReleaseDC (GDK_DRAWABLE_IMPL_WIN32_HANDLE (drawable), hdc);
-    }
-
   return gdk_region_rectangle (&result_rect);
 }
 
@@ -365,6 +368,7 @@ RegisterGdkClass (GdkWindowType wtype)
   static ATOM klassCHILD    = 0;
   static ATOM klassTEMP     = 0;
   static HICON hAppIcon = NULL;
+  static HICON hAppIconSm = NULL;
   static WNDCLASSEX wcl; 
   ATOM klass = 0;
 
@@ -377,26 +381,37 @@ RegisterGdkClass (GdkWindowType wtype)
   wcl.cbWndExtra = 0;
   wcl.hInstance = _gdk_app_hmodule;
   wcl.hIcon = 0;
+  wcl.hIconSm = 0;
   /* initialize once! */
-  if (0 == hAppIcon)
+  if (0 == hAppIcon && 0 == hAppIconSm)
     {
       gchar sLoc [MAX_PATH+1];
 
       if (0 != GetModuleFileName (_gdk_app_hmodule, sLoc, MAX_PATH))
-	{
-	  hAppIcon = ExtractIcon (_gdk_app_hmodule, sLoc, 0);
-	  if (0 == hAppIcon)
-	    {
-	      if (0 != GetModuleFileName (_gdk_dll_hinstance, sLoc, MAX_PATH))
-		hAppIcon = ExtractIcon (_gdk_dll_hinstance, sLoc, 0);
-	    }
-	}
-      if (0 == hAppIcon) 
-	hAppIcon = LoadIcon (NULL, IDI_APPLICATION);
+        {
+          ExtractIconEx (sLoc, 0, &hAppIcon, &hAppIconSm, 1);
+          if (0 == hAppIcon && 0 == hAppIconSm)
+            {
+              if (0 != GetModuleFileName (_gdk_dll_hinstance, sLoc, MAX_PATH))
+                ExtractIconEx (sLoc, 0, &hAppIcon, &hAppIconSm, 1);
+            }
+        }
+      if (0 == hAppIcon && 0 == hAppIconSm)
+        {
+          hAppIcon = LoadImage (NULL, IDI_APPLICATION, IMAGE_ICON,
+                                GetSystemMetrics (SM_CXICON),
+                                GetSystemMetrics (SM_CYICON), 0);
+          hAppIconSm = LoadImage (NULL, IDI_APPLICATION, IMAGE_ICON,
+                                  GetSystemMetrics (SM_CXSMICON),
+                                  GetSystemMetrics (SM_CYSMICON), 0);
+        }
     }
+  if (0 == hAppIcon)
+    hAppIcon = hAppIconSm;
+  else if (0 == hAppIconSm)
+    hAppIconSm = hAppIcon;
 
   wcl.lpszMenuName = NULL;
-  wcl.hIconSm = 0;
 
   /* initialize once per class */
   /*
@@ -406,7 +421,7 @@ RegisterGdkClass (GdkWindowType wtype)
    */
 #define ONCE_PER_CLASS() \
   wcl.hIcon = CopyIcon (hAppIcon); \
-  wcl.hIconSm = CopyIcon (hAppIcon); \
+  wcl.hIconSm = CopyIcon (hAppIconSm); \
   wcl.hbrBackground = NULL; \
   wcl.hCursor = LoadCursor (NULL, IDC_ARROW); 
   
@@ -1069,7 +1084,14 @@ show_window_internal (GdkWindow *window,
 		      SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
       else if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_TOPLEVEL ||
 	       GDK_WINDOW_TYPE (window) == GDK_WINDOW_DIALOG)
-        SetForegroundWindow (GDK_WINDOW_HWND (window));
+	{
+          if (focus_on_map && private->accept_focus)
+            SetForegroundWindow (GDK_WINDOW_HWND (window));
+	  else
+            SetWindowPos (GDK_WINDOW_HWND (window), HWND_TOP,
+	  	          0, 0, 0, 0,
+		          SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
+	}
       else
         BringWindowToTop (GDK_WINDOW_HWND (window));
     }
@@ -1468,7 +1490,16 @@ gdk_window_raise (GdkWindow *window)
       GDK_NOTE (MISC, g_print ("gdk_window_raise: %p\n",
 			       GDK_WINDOW_HWND (window)));
 
-      API_CALL (BringWindowToTop, (GDK_WINDOW_HWND (window)));
+      if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_TEMP)
+        API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), HWND_TOPMOST,
+	                         0, 0, 0, 0,
+				 SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE));
+      else if (((GdkWindowObject *)window)->accept_focus)
+        API_CALL (BringWindowToTop, (GDK_WINDOW_HWND (window)));
+      else
+        API_CALL (SetWindowPos, (GDK_WINDOW_HWND (window), HWND_TOP,
+  			         0, 0, 0, 0,
+			         SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE));
     }
 }
 
@@ -1542,6 +1573,8 @@ gdk_window_set_urgency_hint (GdkWindow *window,
 			     gboolean   urgent)
 {
   FLASHWINFO flashwinfo;
+  typedef BOOL (*PFN_FlashWindowEx) (FLASHWINFO*);
+  PFN_FlashWindowEx flashWindowEx = NULL;
 
   g_return_if_fail (GDK_IS_WINDOW (window));
   g_return_if_fail (GDK_WINDOW_TYPE (window) != GDK_WINDOW_CHILD);
@@ -1549,16 +1582,106 @@ gdk_window_set_urgency_hint (GdkWindow *window,
   if (GDK_WINDOW_DESTROYED (window))
     return;
 
-  flashwinfo.cbSize = sizeof (flashwinfo);
-  flashwinfo.hwnd = GDK_WINDOW_HWND (window);
-  if (urgent)
-    flashwinfo.dwFlags = FLASHW_ALL | FLASHW_TIMER;
-  else
-    flashwinfo.dwFlags = FLASHW_STOP;
-  flashwinfo.uCount = 0;
-  flashwinfo.dwTimeout = 0;
+  flashWindowEx = (PFN_FlashWindowEx) GetProcAddress (GetModuleHandle ("user32.dll"), "FlashWindowEx");
 
-  FlashWindowEx (&flashwinfo);
+  if (flashWindowEx)
+    {
+      flashwinfo.cbSize = sizeof (flashwinfo);
+      flashwinfo.hwnd = GDK_WINDOW_HWND (window);
+      if (urgent)
+	flashwinfo.dwFlags = FLASHW_ALL | FLASHW_TIMER;
+      else
+	flashwinfo.dwFlags = FLASHW_STOP;
+      flashwinfo.uCount = 0;
+      flashwinfo.dwTimeout = 0;
+      
+      flashWindowEx (&flashwinfo);
+    }
+  else
+    {
+      FlashWindow (GDK_WINDOW_HWND (window), urgent);
+    }
+}
+
+static gboolean
+get_effective_window_decorations (GdkWindow       *window,
+                                  GdkWMDecoration *decoration)
+{
+  GdkWindowImplWin32 *impl;
+
+  impl = (GdkWindowImplWin32 *)((GdkWindowObject *)window)->impl;
+
+  if (gdk_window_get_decorations (window, decoration))
+    return TRUE;
+    
+  if (((GdkWindowObject *) window)->window_type != GDK_WINDOW_TOPLEVEL &&
+      ((GdkWindowObject *) window)->window_type != GDK_WINDOW_DIALOG)
+    return FALSE;
+
+  if ((impl->hint_flags & GDK_HINT_MIN_SIZE) &&
+      (impl->hint_flags & GDK_HINT_MAX_SIZE) &&
+      impl->hints.min_width == impl->hints.max_width &&
+      impl->hints.min_height == impl->hints.max_height)
+    {
+      *decoration = GDK_DECOR_ALL | GDK_DECOR_RESIZEH | GDK_DECOR_MAXIMIZE;
+      if (impl->type_hint == GDK_WINDOW_TYPE_HINT_DIALOG ||
+	  impl->type_hint == GDK_WINDOW_TYPE_HINT_MENU ||
+	  impl->type_hint == GDK_WINDOW_TYPE_HINT_TOOLBAR)
+	*decoration |= GDK_DECOR_MINIMIZE;
+      else if (impl->type_hint == GDK_WINDOW_TYPE_HINT_SPLASHSCREEN)
+	*decoration |= GDK_DECOR_MENU | GDK_DECOR_MINIMIZE;
+
+      return TRUE;
+    }
+  else if (impl->hint_flags & GDK_HINT_MAX_SIZE)
+    {
+      *decoration = GDK_DECOR_ALL | GDK_DECOR_MAXIMIZE;
+      if (impl->type_hint == GDK_WINDOW_TYPE_HINT_DIALOG ||
+	  impl->type_hint == GDK_WINDOW_TYPE_HINT_MENU ||
+	  impl->type_hint == GDK_WINDOW_TYPE_HINT_TOOLBAR)
+	*decoration |= GDK_DECOR_MINIMIZE;
+      return TRUE;
+    }
+  else
+    {
+      switch (impl->type_hint)
+	{
+	case GDK_WINDOW_TYPE_HINT_DIALOG:
+	  *decoration = (GDK_DECOR_ALL | GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE);
+	  return TRUE;
+
+	case GDK_WINDOW_TYPE_HINT_MENU:
+	  *decoration = (GDK_DECOR_ALL | GDK_DECOR_RESIZEH | GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE);
+	  return TRUE;
+
+	case GDK_WINDOW_TYPE_HINT_TOOLBAR:
+	  gdk_window_set_skip_taskbar_hint (window, TRUE);
+	  *decoration = GDK_DECOR_ALL | GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE;
+	  return TRUE;
+
+	case GDK_WINDOW_TYPE_HINT_UTILITY:
+	  return FALSE;
+
+	case GDK_WINDOW_TYPE_HINT_SPLASHSCREEN:
+	  *decoration = (GDK_DECOR_ALL | GDK_DECOR_RESIZEH | GDK_DECOR_MENU
+                  | GDK_DECOR_MINIMIZE | GDK_DECOR_MAXIMIZE);
+	  return TRUE;
+	  
+	case GDK_WINDOW_TYPE_HINT_DOCK:
+	  return FALSE;
+	  
+	case GDK_WINDOW_TYPE_HINT_DESKTOP:
+	  return FALSE;
+
+	default:
+	  /* Fall thru */
+	case GDK_WINDOW_TYPE_HINT_NORMAL:
+	  *decoration = GDK_DECOR_ALL;
+	  return TRUE;
+	}
+    }
+    
+  return FALSE;
 }
 
 void 
@@ -1596,27 +1719,6 @@ gdk_window_set_geometry_hints (GdkWindow      *window,
 			       geometry->max_width, geometry->max_height));
     }
 
-  if ((geom_mask & GDK_HINT_MIN_SIZE) &&
-      (geom_mask & GDK_HINT_MAX_SIZE) &&
-      geometry->min_width == geometry->max_width &&
-      geometry->min_height == geometry->max_height)
-    gdk_window_set_decorations (window,
-				GDK_DECOR_ALL |
-				GDK_DECOR_RESIZEH |
-				GDK_DECOR_MAXIMIZE);
-  else if (geom_mask & GDK_HINT_MAX_SIZE)
-    {
-      gdk_window_set_decorations (window,
-				  GDK_DECOR_ALL |
-				  GDK_DECOR_MAXIMIZE);
-      gdk_window_set_decorations (window,
-				  GDK_DECOR_RESIZEH);
-    }
-  else
-    gdk_window_set_decorations (window,
-				GDK_DECOR_RESIZEH |
-				GDK_DECOR_MAXIMIZE);
-
   if (geom_mask & GDK_HINT_BASE_SIZE)
     {
       GDK_NOTE (MISC, g_print ("... BASE_SIZE: %dx%d\n",
@@ -1639,6 +1741,8 @@ gdk_window_set_geometry_hints (GdkWindow      *window,
     {
       GDK_NOTE (MISC, g_print ("... GRAVITY: %d\n", geometry->win_gravity));
     }
+
+  update_style_bits (window);
 }
 
 void
@@ -2233,6 +2337,30 @@ gdk_window_set_events (GdkWindow   *window,
   GDK_WINDOW_OBJECT (window)->event_mask = GDK_STRUCTURE_MASK | event_mask;
 }
 
+static void
+do_shape_combine_region (GdkWindow *window,
+			 HRGN	    hrgn,
+			 gint       x, gint y)
+{
+  RECT rect;
+
+  GetClientRect (GDK_WINDOW_HWND (window), &rect);
+  _gdk_win32_adjust_client_rect (window, &rect);
+
+  OffsetRgn (hrgn, -rect.left, -rect.top);
+  OffsetRgn (hrgn, x, y);
+
+  /* If this is a top-level window, add the title bar to the region */
+  if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_TOPLEVEL)
+    {
+      HRGN tmp = CreateRectRgn (0, 0, rect.right - rect.left, -rect.top);
+      CombineRgn (hrgn, hrgn, tmp, RGN_OR);
+      DeleteObject (tmp);
+    }
+  
+  SetWindowRgn (GDK_WINDOW_HWND (window), hrgn, TRUE);
+}
+
 void
 gdk_window_shape_combine_mask (GdkWindow *window,
 			       GdkBitmap *mask,
@@ -2253,35 +2381,41 @@ gdk_window_shape_combine_mask (GdkWindow *window,
   else
     {
       HRGN hrgn;
-      RECT rect;
-
-      /* Convert mask bitmap to region */
-      hrgn = _gdk_win32_bitmap_to_hrgn (mask);
 
       GDK_NOTE (MISC, g_print ("gdk_window_shape_combine_mask: %p: %p\n",
 			       GDK_WINDOW_HWND (window),
 			       GDK_WINDOW_HWND (mask)));
 
-      GetClientRect (GDK_WINDOW_HWND (window), &rect);
-      _gdk_win32_adjust_client_rect (window, &rect);
+      /* Convert mask bitmap to region */
+      hrgn = _gdk_win32_bitmap_to_hrgn (mask);
 
-      OffsetRgn (hrgn, -rect.left, -rect.top);
-      OffsetRgn (hrgn, x, y);
-
-      /* If this is a top-level window, add the title bar to the region */
-      if (GDK_WINDOW_TYPE (window) == GDK_WINDOW_TOPLEVEL)
-	{
-	  HRGN tmp = CreateRectRgn (0, 0, rect.right - rect.left, -rect.top);
-	  CombineRgn (hrgn, hrgn, tmp, RGN_OR);
-	  DeleteObject (tmp);
-	}
-      
-      SetWindowRgn (GDK_WINDOW_HWND (window), hrgn, TRUE);
+      do_shape_combine_region (window, hrgn, x, y);
 
       private->shaped = TRUE;
     }
 }
 
+void 
+gdk_window_input_shape_combine_mask (GdkWindow *window,
+				     GdkBitmap *mask,
+				     gint       x,
+				     gint       y)
+{
+  g_return_if_fail (GDK_IS_WINDOW (window));
+
+  /* Not yet implemented
+   *
+   * I don't think there is anything in the Win32 API to directly
+   * support this. And anyway, as we don't currently support RGBA
+   * windows, it doesn't really matter.
+   *
+   * When we do support RGBA, input shape functionality could probably
+   * be implemented by saving the input shape region in the per-window
+   * private data, and then simply checking before generating an input
+   * event whether the event's coordinates are inside the region.
+   */
+}
+
 void
 gdk_window_set_override_redirect (GdkWindow *window,
 				  gboolean   override_redirect)
@@ -2466,11 +2600,27 @@ gdk_window_set_group (GdkWindow *window,
 }
 
 static void
-set_or_clear_style_bits (GdkWindow *window,
-			 gboolean   clear_bits,
-			 int        bits)
+update_single_bit (LONG *style,
+                   BOOL  all,
+		   int   gdk_bit,
+		   int   style_bit)
 {
+  /* all controls the interpretation of gdk_bit -- if all is true, gdk_bit
+     indicates whether style_bit is off; if all is false, gdk bit indicate whether
+     style_bit is on */
+  if ((!all && gdk_bit) || (all && !gdk_bit))
+    *style |= style_bit;
+  else
+    *style &= ~style_bit;
+}
+
+static void
+update_style_bits (GdkWindow *window)
+{
+  GdkWMDecoration decorations;
+  GdkWMFunction functions;
   LONG style, exstyle;
+  BOOL all;
   RECT rect, before, after;
 
   style = GetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE);
@@ -2480,11 +2630,29 @@ set_or_clear_style_bits (GdkWindow *window,
   after = before;
   AdjustWindowRectEx (&before, style, FALSE, exstyle);
 
-  if (clear_bits)
-    style &= ~bits;
-  else
-    style |= bits;
+  if (get_effective_window_decorations (window, &decorations))
+    {
+      all = (decorations & GDK_DECOR_ALL);
+      update_single_bit (&style, all, decorations & GDK_DECOR_BORDER, WS_BORDER);
+      update_single_bit (&style, all, decorations & GDK_DECOR_RESIZEH, WS_THICKFRAME);
+      update_single_bit (&style, all, decorations & GDK_DECOR_TITLE, WS_CAPTION);
+      update_single_bit (&style, all, decorations & GDK_DECOR_MENU, WS_SYSMENU);
+      update_single_bit (&style, all, decorations & GDK_DECOR_MINIMIZE, WS_MINIMIZEBOX);
+      update_single_bit (&style, all, decorations & GDK_DECOR_MAXIMIZE, WS_MAXIMIZEBOX);
+    }
 
+  /* XXX this is actually incorrect.  The menu entries should be added or removed
+     from the system menu without affecting the window style. */
+  if (_gdk_window_get_functions (window, &functions))
+    {
+      all = (functions & GDK_DECOR_ALL);
+      update_single_bit (&style, all, functions & GDK_FUNC_RESIZE, WS_THICKFRAME);
+      update_single_bit (&style, all, functions & GDK_FUNC_MOVE, WS_THICKFRAME | WS_SYSMENU);
+      update_single_bit (&style, all, functions & GDK_FUNC_MINIMIZE, WS_MINIMIZE);
+      update_single_bit (&style, all, functions & GDK_FUNC_MOVE, WS_MAXIMIZE);
+      update_single_bit (&style, all, functions & GDK_FUNC_CLOSE, WS_SYSMENU);
+    }
+    
   SetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE, style);
 
   AdjustWindowRectEx (&after, style, FALSE, exstyle);
@@ -2502,12 +2670,23 @@ set_or_clear_style_bits (GdkWindow *window,
 		SWP_NOREPOSITION | SWP_NOZORDER);
 }
 
+static GQuark
+get_decorations_quark ()
+{
+  static GQuark quark = 0;
+  
+  if (!quark)
+    quark = g_quark_from_static_string ("gdk-window-decorations");
+  
+  return quark;
+}
+
 void
 gdk_window_set_decorations (GdkWindow      *window,
 			    GdkWMDecoration decorations)
 {
-  int bits;
-
+  GdkWMDecoration* decorations_copy;
+  
   g_return_if_fail (GDK_IS_WINDOW (window));
   
   GDK_NOTE (MISC, g_print ("gdk_window_set_decorations: %p: %s %s%s%s%s%s%s\n",
@@ -2520,56 +2699,44 @@ gdk_window_set_decorations (GdkWindow      *window,
 			   (decorations & GDK_DECOR_MINIMIZE ? "MINIMIZE " : ""),
 			   (decorations & GDK_DECOR_MAXIMIZE ? "MAXIMIZE " : "")));
 
-  bits = 0;
+  decorations_copy = g_malloc (sizeof (GdkWMDecoration));
+  *decorations_copy = decorations;
+  g_object_set_qdata_full (G_OBJECT (window), get_decorations_quark (), decorations_copy, g_free);
 
-  if (decorations & GDK_DECOR_BORDER)
-    bits |= WS_BORDER;
-  if (decorations & GDK_DECOR_RESIZEH)
-    bits |= WS_THICKFRAME;
-  if (decorations & GDK_DECOR_TITLE)
-    bits |= WS_CAPTION;
-  if (decorations & GDK_DECOR_MENU)
-    bits |= WS_SYSMENU;
-  if (decorations & GDK_DECOR_MINIMIZE)
-    bits |= WS_MINIMIZEBOX;
-  if (decorations & GDK_DECOR_MAXIMIZE)
-    bits |= WS_MAXIMIZEBOX;
-
-  set_or_clear_style_bits (window, (decorations & GDK_DECOR_ALL), bits);
+  update_style_bits (window);
 }
 
 gboolean
 gdk_window_get_decorations (GdkWindow       *window,
 			    GdkWMDecoration *decorations)
 {
-  LONG style;
-
+  GdkWMDecoration* decorations_set;
+  
   g_return_val_if_fail (GDK_IS_WINDOW (window), FALSE);
 
-  style = GetWindowLong (GDK_WINDOW_HWND (window), GWL_STYLE);
-  *decorations = 0;
+  decorations_set = g_object_get_qdata (G_OBJECT (window), get_decorations_quark ());
+  if (decorations_set)
+    *decorations = *decorations_set;
 
-  if (style & WS_BORDER)
-    *decorations |= GDK_DECOR_BORDER;
-  if (style & WS_THICKFRAME)
-    *decorations |= GDK_DECOR_RESIZEH;
-  if (style & WS_CAPTION)
-    *decorations |= GDK_DECOR_TITLE;
-  if (style & WS_SYSMENU)
-    *decorations |= GDK_DECOR_MENU;
-  if (style & WS_MINIMIZEBOX)
-    *decorations |= GDK_DECOR_MINIMIZE;
-  if (style & WS_MAXIMIZEBOX)
-    *decorations |= GDK_DECOR_MAXIMIZE;
+  return (decorations_set != NULL);
+}
 
-  return *decorations != 0;
+static GQuark
+get_functions_quark ()
+{
+  static GQuark quark = 0;
+  
+  if (!quark)
+    quark = g_quark_from_static_string ("gdk-window-functions");
+  
+  return quark;
 }
 
 void
 gdk_window_set_functions (GdkWindow    *window,
 			  GdkWMFunction functions)
 {
-  int bits;
+  GdkWMFunction* functions_copy;
 
   g_return_if_fail (GDK_IS_WINDOW (window));
   
@@ -2582,20 +2749,24 @@ gdk_window_set_functions (GdkWindow    *window,
 			   (functions & GDK_FUNC_MAXIMIZE ? "MAXIMIZE " : ""),
 			   (functions & GDK_FUNC_CLOSE ? "CLOSE " : "")));
 
-  bits = 0;
+  functions_copy = g_malloc (sizeof (GdkWMFunction));
+  *functions_copy = functions;
+  g_object_set_qdata_full (G_OBJECT (window), get_functions_quark (), functions_copy, g_free);
 
-  if (functions & GDK_FUNC_RESIZE)
-    bits |= WS_THICKFRAME;
-  if (functions & GDK_FUNC_MOVE)
-    bits |= (WS_THICKFRAME|WS_SYSMENU);
-  if (functions & GDK_FUNC_MINIMIZE)
-    bits |= WS_MINIMIZEBOX;
-  if (functions & GDK_FUNC_MAXIMIZE)
-    bits |= WS_MAXIMIZEBOX;
-  if (functions & GDK_FUNC_CLOSE)
-    bits |= WS_SYSMENU;
+  update_style_bits (window);
+}
+
+gboolean
+_gdk_window_get_functions (GdkWindow     *window,
+		           GdkWMFunction *functions)
+{
+  GdkWMDecoration* functions_set;
   
-  set_or_clear_style_bits (window, (functions & GDK_FUNC_ALL), bits);
+  functions_set = g_object_get_qdata (G_OBJECT (window), get_functions_quark ());
+  if (functions_set)
+    *functions = *functions_set;
+
+  return (functions_set != NULL);
 }
 
 static void
@@ -2691,6 +2862,26 @@ gdk_window_merge_child_shapes (GdkWindow *window)
   gdk_propagate_shapes (GDK_WINDOW_HWND (window), TRUE);
 }
 
+void 
+gdk_window_set_child_input_shapes (GdkWindow *window)
+{
+  g_return_if_fail (GDK_IS_WINDOW (window));
+  
+  /* Not yet implemented. See comment in
+   * gdk_window_input_shape_combine_mask().
+   */
+}
+
+void 
+gdk_window_merge_child_input_shapes (GdkWindow *window)
+{
+  g_return_if_fail (GDK_IS_WINDOW (window));
+  
+  /* Not yet implemented. See comment in
+   * gdk_window_input_shape_combine_mask().
+   */
+}
+
 gboolean 
 gdk_window_set_static_gravities (GdkWindow *window,
 				 gboolean   use_static)
@@ -3135,40 +3326,21 @@ gdk_window_set_type_hint (GdkWindow        *window,
 
   GDK_NOTE (MISC, g_print ("gdk_window_set_type_hint: %p: %d\n",
 			   GDK_WINDOW_HWND (window), hint));
-  switch (hint)
-    {
-    case GDK_WINDOW_TYPE_HINT_DIALOG:
-      break;
-    case GDK_WINDOW_TYPE_HINT_MENU:
-      gdk_window_set_decorations (window,
-				  GDK_DECOR_ALL |
-				  GDK_DECOR_RESIZEH |
-				  GDK_DECOR_MINIMIZE |
-				  GDK_DECOR_MAXIMIZE);
-      break;
-    case GDK_WINDOW_TYPE_HINT_TOOLBAR:
-      gdk_window_set_skip_taskbar_hint (window, TRUE);
-      break;
-    case GDK_WINDOW_TYPE_HINT_UTILITY:
-      break;
-    case GDK_WINDOW_TYPE_HINT_SPLASHSCREEN:
-      gdk_window_set_decorations (window,
-				  GDK_DECOR_ALL |
-				  GDK_DECOR_RESIZEH |
-				  GDK_DECOR_MENU |
-				  GDK_DECOR_MINIMIZE |
-				  GDK_DECOR_MAXIMIZE);
-      break;
-    case GDK_WINDOW_TYPE_HINT_DOCK:
-      break;
-    case GDK_WINDOW_TYPE_HINT_DESKTOP:
-      break;
-    default:
-      g_warning ("Unknown hint %d passed to gdk_window_set_type_hint", hint);
-      /* Fall thru */
-    case GDK_WINDOW_TYPE_HINT_NORMAL:
-      break;
-    }
+
+  ((GdkWindowImplWin32 *)((GdkWindowObject *)window)->impl)->type_hint = hint;
+
+  update_style_bits (window);
+}
+
+GdkWindowTypeHint
+gdk_window_get_type_hint (GdkWindow *window)
+{
+  g_return_val_if_fail (GDK_IS_WINDOW (window), GDK_WINDOW_TYPE_HINT_NORMAL);
+  
+  if (GDK_WINDOW_DESTROYED (window))
+    return GDK_WINDOW_TYPE_HINT_NORMAL;
+
+  return GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->type_hint;
 }
 
 void
@@ -3177,12 +3349,48 @@ gdk_window_shape_combine_region (GdkWindow *window,
                                  gint       offset_x,
                                  gint       offset_y)
 {
+  GdkWindowObject *private = (GdkWindowObject *)window;
+
   g_return_if_fail (GDK_IS_WINDOW (window));
 
   if (GDK_WINDOW_DESTROYED (window))
     return;
 
-  /* XXX: even on X implemented conditional ... */  
+  if (!shape_region)
+    {
+      GDK_NOTE (MISC, g_print ("gdk_window_shape_combine_region: %p: none\n",
+			       GDK_WINDOW_HWND (window)));
+      SetWindowRgn (GDK_WINDOW_HWND (window), NULL, TRUE);
+
+      private->shaped = FALSE;
+    }
+  else
+    {
+      HRGN hrgn;
+
+      hrgn = _gdk_win32_gdkregion_to_hrgn (shape_region, 0, 0);
+      
+      GDK_NOTE (MISC, g_print ("gdk_window_shape_combine_region: %p: %p\n",
+			       GDK_WINDOW_HWND (window),
+			       hrgn));
+
+      do_shape_combine_region (window, hrgn, offset_x, offset_y);
+
+      private->shaped = TRUE;
+    }
+}
+
+void 
+gdk_window_input_shape_combine_region (GdkWindow *window,
+				       GdkRegion *shape_region,
+				       gint       offset_x,
+				       gint       offset_y)
+{
+  g_return_if_fail (GDK_IS_WINDOW (window));
+
+  /* Not yet implemented. See comment in
+   * gdk_window_input_shape_combine_mask().
+   */
 }
 
 GdkWindow *
@@ -3196,9 +3404,11 @@ gdk_window_lookup_for_display (GdkDisplay *display, GdkNativeWindow anid)
 void
 gdk_window_enable_synchronized_configure (GdkWindow *window)
 {
+  g_return_if_fail (GDK_IS_WINDOW (window));
 }
 
 void
 gdk_window_configure_finished (GdkWindow *window)
 {
+  g_return_if_fail (GDK_IS_WINDOW (window));
 }
