@@ -698,6 +698,30 @@ gtk_calendar_init (GtkCalendar *calendar)
   else if (strcmp (year_before, "calendar:MY") != 0)
     g_warning ("Whoever translated calendar:MY did so wrongly.\n");
 
+#ifdef G_OS_WIN32
+  priv->week_start = 0;
+  week_start = NULL;
+
+  if (G_WIN32_HAVE_WIDECHAR_API ())
+    {
+      if (GetLocaleInfoW (GetThreadLocale (), LOCALE_IFIRSTDAYOFWEEK,
+			  wbuffer, G_N_ELEMENTS (wbuffer)))
+	week_start = g_utf16_to_utf8 (wbuffer, -1, NULL, NULL, NULL);
+    }
+  else
+    {
+      if (GetLocaleInfoA (GetThreadLocale (),
+			  LOCALE_IFIRSTDAYOFWEEK | LOCALE_USE_CP_ACP,
+			  buffer, G_N_ELEMENTS (buffer)))
+	week_start = g_locale_to_utf8 (buffer, -1, NULL, NULL, NULL);
+    }
+  
+  if (week_start != NULL)
+    {
+      priv->week_start = (week_start[0] - '0' + 1) % 7;
+      g_free(week_start);
+    }
+#else
 #ifdef HAVE__NL_TIME_FIRST_WEEKDAY
   langinfo = nl_langinfo (_NL_TIME_FIRST_WEEKDAY);
   first_weekday = langinfo[0];
@@ -728,6 +752,7 @@ gtk_calendar_init (GtkCalendar *calendar)
       g_warning ("Whoever translated calendar:week_start:0 did so wrongly.\n");
       priv->week_start = 0;
     }
+#endif
 #endif
 
   calendar_compute_days (calendar);
@@ -2309,7 +2334,7 @@ calendar_paint_arrow (GtkCalendar *calendar,
 	gtk_paint_arrow (widget->style, window, state, 
 			 GTK_SHADOW_OUT, NULL, widget, "calendar",
 			 GTK_ARROW_RIGHT, TRUE, 
-			 width/2 - 2, height/2 - 4, 8, 8);
+			 width/2 - 4, height/2 - 4, 8, 8);
     }
 }
 
