@@ -87,6 +87,7 @@ static void     update_collate_icon                (GtkToggleButton    *toggle_b
 						    GtkPrintUnixDialog *dialog);
 static gboolean dialog_get_collate                 (GtkPrintUnixDialog *dialog);
 static gboolean dialog_get_reverse                 (GtkPrintUnixDialog *dialog);
+static gint     dialog_get_n_copies                (GtkPrintUnixDialog *dialog);
 
 enum {
   PROP_0,
@@ -1018,8 +1019,10 @@ update_dialog_from_capabilities (GtkPrintUnixDialog *dialog)
   GtkPrintCapabilities caps;
   GtkPrintUnixDialogPrivate *priv = dialog->priv;
   gboolean can_collate;
+  const gchar *copies;
 
-  can_collate = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (priv->copies_spin)) > 1;
+  copies = gtk_entry_get_text (GTK_ENTRY (priv->copies_spin));
+  can_collate = (*copies != '\0' && atoi (copies) > 1);
 
   caps = priv->manual_capabilities | priv->printer_capabilities;
 
@@ -1033,7 +1036,7 @@ update_dialog_from_capabilities (GtkPrintUnixDialog *dialog)
   gtk_widget_set_sensitive (priv->reverse_check,
 			    caps & GTK_PRINT_CAPABILITY_REVERSE);
   gtk_widget_set_sensitive (priv->scale_spin,
-			    caps & GTK_PRINT_CAPABILITY_PAGE_SET);
+			    caps & GTK_PRINT_CAPABILITY_SCALE);
 
   if (caps & GTK_PRINT_CAPABILITY_PREVIEW)
     gtk_widget_show (priv->preview_button);
@@ -1331,10 +1334,13 @@ draw_collate_cb (GtkWidget	    *widget,
   gint size;
   gfloat scale;
   gboolean collate, reverse, rtl;
+  gint copies;
   gint text_x;
 
   collate = dialog_get_collate (dialog);
   reverse = dialog_get_reverse (dialog);
+  copies = dialog_get_n_copies (dialog);
+
   rtl = (gtk_widget_get_direction (GTK_WIDGET (widget)) == GTK_TEXT_DIR_RTL);
 
   settings = gtk_widget_get_settings (widget);
@@ -1347,12 +1353,20 @@ draw_collate_cb (GtkWidget	    *widget,
 
   cr = gdk_cairo_create (widget->window);
 
-  paint_page (widget, cr, scale, rtl ? 40: 15, 5, collate == reverse ? "1" : "2", text_x);
-  paint_page (widget, cr, scale, rtl ? 50: 5, 15, reverse ? "2" : "1", text_x);
+  if (copies == 1)
+    {
+      paint_page (widget, cr, scale, rtl ? 40: 15, 5, reverse ? "1" : "2", text_x);
+      paint_page (widget, cr, scale, rtl ? 50: 5, 15, reverse ? "2" : "1", text_x);
+    }
+  else
+    {
+      paint_page (widget, cr, scale, rtl ? 40: 15, 5, collate == reverse ? "1" : "2", text_x);
+      paint_page (widget, cr, scale, rtl ? 50: 5, 15, reverse ? "2" : "1", text_x);
+
+      paint_page (widget, cr, scale, rtl ? 5 : 50, 5, reverse ? "1" : "2", text_x);
+      paint_page (widget, cr, scale, rtl ? 15 : 40, 15, collate == reverse ? "2" : "1", text_x);
+    }
 
-  paint_page (widget, cr, scale, rtl ? 5 : 50, 5, reverse ? "1" : "2", text_x);
-  paint_page (widget, cr, scale, rtl ? 15 : 40, 15, collate == reverse ? "2" : "1", text_x);
-  
   cairo_destroy (cr);
 
   return TRUE;
@@ -1542,6 +1556,8 @@ create_main_page (GtkPrintUnixDialog *dialog)
   gtk_label_set_mnemonic_widget (GTK_LABEL (label), spinbutton);
   g_signal_connect_swapped (spinbutton, "value-changed", 
 			    G_CALLBACK (update_dialog_from_capabilities), dialog);
+  g_signal_connect_swapped (spinbutton, "changed", 
+			    G_CALLBACK (update_dialog_from_capabilities), dialog);
   
   check = gtk_check_button_new_with_mnemonic (_("C_ollate"));
   priv->collate_check = check;
@@ -2732,8 +2748,23 @@ void
 gtk_print_unix_dialog_set_manual_capabilities (GtkPrintUnixDialog   *dialog,
 					       GtkPrintCapabilities  capabilities)
 {
-  dialog->priv->manual_capabilities = capabilities;
+  GtkPrintUnixDialogPrivate *priv = dialog->priv;
+
+  priv->manual_capabilities = capabilities;
   update_dialog_from_capabilities (dialog);
+
+  if (priv->current_printer)
+    {
+      GtkTreeSelection *selection;
+
+      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->printer_treeview));
+
+      g_object_unref (priv->current_printer);
+      priv->current_printer = NULL;
+      priv->internal_printer_change = TRUE;
+      selected_printer_changed (selection, dialog);
+      priv->internal_printer_change = FALSE;
+   }
 }
 
 #define __GTK_PRINT_UNIX_DIALOG_C__
