@@ -32,12 +32,17 @@
 #include "x11/gdkx.h"
 #endif
 
+#define DEFAULT_TIMEOUT_INITIAL 200
+#define DEFAULT_TIMEOUT_REPEAT   20
+#define DEFAULT_TIMEOUT_EXPAND  500
+
 typedef struct _GtkSettingsValuePrivate GtkSettingsValuePrivate;
 
 typedef enum
 {
   GTK_SETTINGS_SOURCE_DEFAULT,
   GTK_SETTINGS_SOURCE_RC_FILE,
+  GTK_SETTINGS_SOURCE_XSETTING,
   GTK_SETTINGS_SOURCE_APPLICATION
 } GtkSettingsSource;
 
@@ -78,13 +83,20 @@ enum {
   PROP_CURSOR_THEME_NAME,
   PROP_CURSOR_THEME_SIZE,
 #endif
-  PROP_ALTERNATIVE_BUTTON_ORDER
+  PROP_ALTERNATIVE_BUTTON_ORDER,
+  PROP_SHOW_INPUT_METHOD_MENU,
+  PROP_SHOW_UNICODE_MENU,
+  PROP_TIMEOUT_INITIAL,
+  PROP_TIMEOUT_REPEAT,
+  PROP_TIMEOUT_EXPAND,
+  PROP_COLOR_SCHEME,
+  PROP_ENABLE_ANIMATIONS,
+  PROP_TOUCHSCREEN_MODE,
+  PROP_COLOR_HASH
 };
 
 
 /* --- prototypes --- */
-static void	gtk_settings_init		 (GtkSettings		*settings);
-static void	gtk_settings_class_init		 (GtkSettingsClass	*class);
 static void	gtk_settings_finalize		 (GObject		*object);
 static void	gtk_settings_get_property	 (GObject		*object,
 						  guint			 property_id,
@@ -107,44 +119,24 @@ static void    settings_update_cursor_theme      (GtkSettings           *setting
 static void    settings_update_resolution        (GtkSettings           *settings);
 static void    settings_update_font_options      (GtkSettings           *settings);
 #endif
+static void    settings_update_color_scheme      (GtkSettings *settings);
 
+static void    merge_color_scheme                (GtkSettings           *settings, 
+						  GValue                *value, 
+						  GtkSettingsSource      source);
+static gchar  *get_color_scheme                  (GtkSettings           *settings);
+static GHashTable *get_color_hash                (GtkSettings           *settings);
 
 
 /* --- variables --- */
-static gpointer		 parent_class = NULL;
 static GQuark		 quark_property_parser = 0;
 static GSList           *object_list = NULL;
 static guint		 class_n_properties = 0;
 
 
-/* --- functions --- */
-GType
-gtk_settings_get_type (void)
-{
-  static GType settings_type = 0;
-  
-  if (!settings_type)
-    {
-      static const GTypeInfo settings_info =
-      {
-	sizeof (GtkSettingsClass),
-	NULL,           /* base_init */
-	NULL,           /* base_finalize */
-	(GClassInitFunc) gtk_settings_class_init,
-	NULL,           /* class_finalize */
-	NULL,           /* class_data */
-	sizeof (GtkSettings),
-	0,              /* n_preallocs */
-	(GInstanceInitFunc) gtk_settings_init,
-      };
-      
-      settings_type = g_type_register_static (G_TYPE_OBJECT, "GtkSettings",
-					      &settings_info, 0);
-    }
-  
-  return settings_type;
-}
+G_DEFINE_TYPE (GtkSettings, gtk_settings, G_TYPE_OBJECT)
 
+/* --- functions --- */
 static void
 gtk_settings_init (GtkSettings *settings)
 {
@@ -187,15 +179,12 @@ gtk_settings_class_init (GtkSettingsClass *class)
   GObjectClass *gobject_class = G_OBJECT_CLASS (class);
   guint result;
   
-  parent_class = g_type_class_peek_parent (class);
-
   gobject_class->finalize = gtk_settings_finalize;
   gobject_class->get_property = gtk_settings_get_property;
   gobject_class->set_property = gtk_settings_set_property;
   gobject_class->notify = gtk_settings_notify;
 
   quark_property_parser = g_quark_from_static_string ("gtk-rc-property-parser");
-
   result = settings_install_property_parser (class,
                                              g_param_spec_int ("gtk-double-click-time",
                                                                P_("Double Click Time"),
@@ -371,7 +360,7 @@ gtk_settings_class_init (GtkSettingsClass *class)
   result = settings_install_property_parser (class,
                                              g_param_spec_string ("gtk-cursor-theme-name",
 								  P_("Cursor theme name"),
-								  P_("Name of the cursor theme to use"),
+								  P_("Name of the cursor theme to use, or NULL to use the default theme"),
 								  NULL,
 								  GTK_PARAM_READWRITE),
                                              NULL);
@@ -380,8 +369,8 @@ gtk_settings_class_init (GtkSettingsClass *class)
   result = settings_install_property_parser (class,
 					     g_param_spec_int ("gtk-cursor-theme-size",
  							       P_("Cursor theme size"),
- 							       P_("Size to use for cursors"),
- 							       0, 128, 24,
+ 							       P_("Size to use for cursors, or 0 to use the default size"),
+ 							       0, 128, 0,
  							       GTK_PARAM_READWRITE),
 					     NULL);
   
@@ -396,6 +385,131 @@ gtk_settings_class_init (GtkSettingsClass *class)
 								   GTK_PARAM_READWRITE),
                                              NULL);
   g_assert (result == PROP_ALTERNATIVE_BUTTON_ORDER);
+
+  result = settings_install_property_parser (class,
+					     g_param_spec_boolean ("gtk-show-input-method-menu",
+								   P_("Show the 'Input Methods' menu"),
+								   P_("Whether the context menus of entries and text views should offer to change the input method"),
+								   TRUE,
+								   GTK_PARAM_READWRITE),
+					     NULL);
+  g_assert (result == PROP_SHOW_INPUT_METHOD_MENU);
+
+  result = settings_install_property_parser (class,
+					     g_param_spec_boolean ("gtk-show-unicode-menu",
+								   P_("Show the 'Insert Unicode Control Character' menu"),
+								   P_("Whether the context menus of entries and text views should offer to insert control characters"),
+								   TRUE,
+								   GTK_PARAM_READWRITE),
+					     NULL);
+  g_assert (result == PROP_SHOW_UNICODE_MENU);
+
+  result = settings_install_property_parser (class,
+					     g_param_spec_int ("gtk-timeout-initial",
+ 							       P_("Start timeout"),
+ 							       P_("Starting value for timeouts, when button is pressed"),
+ 							       0, G_MAXINT, DEFAULT_TIMEOUT_INITIAL,
+ 							       GTK_PARAM_READWRITE),
+					     NULL);
+
+  g_assert (result == PROP_TIMEOUT_INITIAL);
+
+  result = settings_install_property_parser (class,
+					     g_param_spec_int ("gtk-timeout-repeat",
+ 							       P_("Repeat timeout"),
+ 							       P_("Repeat value for timeouts, when button is pressed"),
+ 							       0, G_MAXINT, DEFAULT_TIMEOUT_REPEAT,
+ 							       GTK_PARAM_READWRITE),
+					     NULL);
+
+  g_assert (result == PROP_TIMEOUT_REPEAT);
+
+  result = settings_install_property_parser (class,
+					     g_param_spec_int ("gtk-timeout-expand",
+ 							       P_("Expand timeout"),
+ 							       P_("Expand value for timeouts, when a widget is expanding a new region"),
+ 							       0, G_MAXINT, DEFAULT_TIMEOUT_EXPAND,
+ 							       GTK_PARAM_READWRITE),
+					     NULL);
+
+  g_assert (result == PROP_TIMEOUT_EXPAND);
+
+  /**
+   * GtkSettings:gtk-color-scheme:
+   *
+   * A palette of named colors for use in themes. The format of the string is
+   * <programlisting>
+   * name1: color1
+   * name2: color2
+   * ...
+   * </programlisting>
+   * Color names must be acceptable as identifiers in the 
+   * <link linkend="gtk-Resource-Files">gtkrc</link> syntax, and
+   * color specifications must be in the format accepted by
+   * gdk_color_parse().
+   * 
+   * Note that due to the way the color tables from different sources are
+   * merged, color specifications will be converted to hexadecimal form
+   * when getting this property.
+   *
+   * Since: 2.10
+   */
+  result = settings_install_property_parser (class,
+					     g_param_spec_string ("gtk-color-scheme",
+ 								  P_("Color scheme"),
+ 								  P_("A palette of named colors for use in themes"),
+ 								  "",
+ 								  GTK_PARAM_READWRITE),
+					     NULL);
+
+  g_assert (result == PROP_COLOR_SCHEME);
+
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-enable-animations",
+                                                                   P_("Enable Animations"),
+                                                                   P_("Whether to enable toolkit-wide animations."),
+                                                                   TRUE,
+                                                                   GTK_PARAM_READWRITE),
+                                             NULL);
+
+  g_assert (result == PROP_ENABLE_ANIMATIONS);
+
+  /**
+   * GtkSettings:gtk-touchscreen-mode:
+   *
+   * When TRUE, there are no motion notify events delivered on this screen,
+   * and widgets can't use the pointer hovering them for any essential
+   * functionality.
+   *
+   * Since: 2.10
+   */
+  result = settings_install_property_parser (class,
+                                             g_param_spec_boolean ("gtk-touchscreen-mode",
+                                                                   P_("Enable Touchscreen Mode"),
+                                                                   P_("When TRUE, there are no motion notify events delivered on this screen"),
+                                                                   FALSE,
+                                                                   GTK_PARAM_READWRITE),
+                                             NULL);
+
+  g_assert (result == PROP_TOUCHSCREEN_MODE);
+
+  /**
+   * GtkSettings:color-hash:
+   *
+   * Holds a hash table representation of the gtk-color-scheme setting,
+   * mapping color names to #GdkColor<!-- -->s. 
+   *
+   * Since: 2.10
+   */
+  g_object_class_install_property (gobject_class,
+				   PROP_COLOR_HASH,
+				   g_param_spec_boxed ("color-hash",
+						       P_("Color Hash"),
+						       P_("A hash table representation of the color scheme."),
+						       G_TYPE_HASH_TABLE,
+						       GTK_PARAM_READABLE));
+
+  class_n_properties++;
 }
 
 static void
@@ -406,13 +520,15 @@ gtk_settings_finalize (GObject *object)
 
   object_list = g_slist_remove (object_list, settings);
 
+  _gtk_rc_context_destroy (settings);
+
   for (i = 0; i < class_n_properties; i++)
     g_value_unset (&settings->property_values[i].value);
   g_free (settings->property_values);
-
+  
   g_datalist_clear (&settings->queued_settings);
 
-  G_OBJECT_CLASS (parent_class)->finalize (object);
+  G_OBJECT_CLASS (gtk_settings_parent_class)->finalize (object);
 }
 
 /**
@@ -437,7 +553,8 @@ gtk_settings_get_for_screen (GdkScreen *screen)
     {
       settings = g_object_new (GTK_TYPE_SETTINGS, NULL);
       settings->screen = screen;
-      g_object_set_data (G_OBJECT (screen), "gtk-settings", settings);
+      g_object_set_data_full (G_OBJECT (screen), I_("gtk-settings"), 
+			      settings, g_object_unref);
 
       gtk_rc_reparse_all_for_settings (settings, TRUE);
       settings_update_double_click (settings);
@@ -446,6 +563,7 @@ gtk_settings_get_for_screen (GdkScreen *screen)
       settings_update_resolution (settings);
       settings_update_font_options (settings);
 #endif
+      settings_update_color_scheme (settings);
     }
   
   return settings;
@@ -493,6 +611,12 @@ gtk_settings_get_property (GObject     *object,
   GType value_type = G_VALUE_TYPE (value);
   GType fundamental_type = G_TYPE_FUNDAMENTAL (value_type);
 
+  if (property_id == PROP_COLOR_HASH)
+    {
+      g_value_set_boxed (value, get_color_hash (settings));
+      return;
+    }
+
   /* For enums and strings, we need to get the value as a string,
    * not as an int, since we support using names/nicks as the setting
    * value.
@@ -505,8 +629,16 @@ gtk_settings_get_property (GObject     *object,
       if (settings->property_values[property_id - 1].source == GTK_SETTINGS_SOURCE_APPLICATION ||
 	  !gdk_screen_get_setting (settings->screen, pspec->name, value))
         g_value_copy (&settings->property_values[property_id - 1].value, value);
-      else
-        g_param_value_validate (pspec, value);
+      else 
+	{
+	  if (pspec->param_id == PROP_COLOR_SCHEME)
+	    {
+	      merge_color_scheme (settings, value, GTK_SETTINGS_SOURCE_XSETTING);
+	      g_value_take_string (value, get_color_scheme (settings));
+	    }
+	  
+	  g_param_value_validate (pspec, value);
+	}
     }
   else
     {
@@ -528,7 +660,6 @@ gtk_settings_get_property (GObject     *object,
           GtkRcPropertyParser parser = (GtkRcPropertyParser) g_param_spec_get_qdata (pspec, quark_property_parser);
           
           g_value_init (&gstring_value, G_TYPE_GSTRING);
-
           g_value_take_boxed (&gstring_value,
                               g_string_new (g_value_get_string (&val)));
 
@@ -572,6 +703,9 @@ gtk_settings_notify (GObject    *object,
     case PROP_DOUBLE_CLICK_DISTANCE:
       settings_update_double_click (settings);
       break;
+    case PROP_COLOR_SCHEME:
+      settings_update_color_scheme (settings);
+      break;
 #ifdef GDK_WINDOWING_X11
     case PROP_XFT_DPI:
       settings_update_resolution (settings);
@@ -675,7 +809,14 @@ apply_queued_setting (GtkSettings             *data,
   if (_gtk_settings_parse_convert (parser, &qvalue->public.value,
 				   pspec, &tmp_value))
     {
-      if (data->property_values[pspec->param_id - 1].source <= qvalue->source)
+      if (pspec->param_id == PROP_COLOR_SCHEME) 
+	{
+	  merge_color_scheme (data, &tmp_value, qvalue->source);
+	  g_object_set_property (G_OBJECT (data), pspec->name, &tmp_value);
+	  data->property_values[pspec->param_id - 1].source = GTK_SETTINGS_SOURCE_DEFAULT;
+	}      
+      
+      else if (data->property_values[pspec->param_id - 1].source <= qvalue->source)
 	{
 	  g_object_set_property (G_OBJECT (data), pspec->name, &tmp_value);
 	  data->property_values[pspec->param_id - 1].source = qvalue->source;
@@ -686,7 +827,7 @@ apply_queued_setting (GtkSettings             *data,
       gchar *debug = g_strdup_value_contents (&qvalue->public.value);
       
       g_message ("%s: failed to retrieve property `%s' of type `%s' from rc file value \"%s\" of type `%s'",
-		 qvalue->public.origin,
+		 qvalue->public.origin ? qvalue->public.origin : "(for origin information, set GTK_DEBUG)",
 		 pspec->name,
 		 g_type_name (G_PARAM_SPEC_VALUE_TYPE (pspec)),
 		 debug,
@@ -833,7 +974,7 @@ gtk_settings_set_property_value_internal (GtkSettings            *settings,
     }
   
   name = g_strdup (prop_name);
-  g_strcanon (name, G_CSET_A_2_Z G_CSET_a_2_z G_CSET_DIGITS "-", '-');
+  g_strcanon (name, G_CSET_DIGITS "-" G_CSET_a_2_z G_CSET_A_2_Z, '-');
   name_quark = g_quark_from_string (name);
   g_free (name);
 
@@ -865,7 +1006,6 @@ gtk_settings_set_property_value (GtkSettings            *settings,
   g_return_if_fail (GTK_SETTINGS (settings));
   g_return_if_fail (prop_name != NULL);
   g_return_if_fail (new_value != NULL);
-  g_return_if_fail (new_value->origin != NULL);
 
   gtk_settings_set_property_value_internal (settings, prop_name, new_value,
 					    GTK_SETTINGS_SOURCE_APPLICATION);
@@ -879,7 +1019,6 @@ _gtk_settings_set_property_value_from_rc (GtkSettings            *settings,
   g_return_if_fail (GTK_SETTINGS (settings));
   g_return_if_fail (prop_name != NULL);
   g_return_if_fail (new_value != NULL);
-  g_return_if_fail (new_value->origin != NULL);
 
   gtk_settings_set_property_value_internal (settings, prop_name, new_value,
 					    GTK_SETTINGS_SOURCE_RC_FILE);
@@ -896,7 +1035,6 @@ gtk_settings_set_string_property (GtkSettings *settings,
   g_return_if_fail (GTK_SETTINGS (settings));
   g_return_if_fail (name != NULL);
   g_return_if_fail (v_string != NULL);
-  g_return_if_fail (origin != NULL);
 
   svalue.origin = (gchar*) origin;
   g_value_init (&svalue.value, G_TYPE_STRING);
@@ -915,7 +1053,6 @@ gtk_settings_set_long_property (GtkSettings *settings,
   
   g_return_if_fail (GTK_SETTINGS (settings));
   g_return_if_fail (name != NULL);
-  g_return_if_fail (origin != NULL);
 
   svalue.origin = (gchar*) origin;
   g_value_init (&svalue.value, G_TYPE_LONG);
@@ -934,7 +1071,6 @@ gtk_settings_set_double_property (GtkSettings *settings,
 
   g_return_if_fail (GTK_SETTINGS (settings));
   g_return_if_fail (name != NULL);
-  g_return_if_fail (origin != NULL);
 
   svalue.origin = (gchar*) origin;
   g_value_init (&svalue.value, G_TYPE_DOUBLE);
@@ -1314,6 +1450,8 @@ _gtk_settings_reset_rc_values (GtkSettings *settings)
       g_datalist_id_remove_data (&settings->queued_settings, key_id);
     }
 
+   g_slist_free (to_reset);
+
   /* Now reset the active settings
    */
   pspecs = g_object_class_list_properties (G_OBJECT_GET_CLASS (settings), NULL);
@@ -1396,8 +1534,6 @@ settings_update_font_options (GtkSettings *settings)
   cairo_antialias_t antialias_mode = CAIRO_ANTIALIAS_DEFAULT;
   gchar *rgba_str;
   cairo_subpixel_order_t subpixel_order = CAIRO_SUBPIXEL_ORDER_DEFAULT;
-  gint dpi_int;
-  double dpi;
   cairo_font_options_t *options;
   
   g_object_get (settings,
@@ -1405,14 +1541,8 @@ settings_update_font_options (GtkSettings *settings)
 		"gtk-xft-hinting", &hinting,
 		"gtk-xft-hintstyle", &hint_style_str,
 		"gtk-xft-rgba", &rgba_str,
-		"gtk-xft-dpi", &dpi_int,
 		NULL);
 
-  if (dpi_int > 0)
-    dpi = dpi_int / 1024.;
-  else
-    dpi = -1.;
-
   options = cairo_font_options_create ();
   
   /* hint_metrics = FALSE should never be set for user interface code.
@@ -1465,7 +1595,7 @@ settings_update_font_options (GtkSettings *settings)
   
   cairo_font_options_set_antialias (options, antialias_mode);
 
-  gdk_screen_set_font_options_libgtk_only (settings->screen, options);
+  gdk_screen_set_font_options (settings->screen, options);
   
   cairo_font_options_destroy (options);
 }
@@ -1485,9 +1615,239 @@ settings_update_resolution (GtkSettings *settings)
   else
     dpi = -1.;
 
-  gdk_screen_set_resolution_libgtk_only (settings->screen, dpi);
+  gdk_screen_set_resolution (settings->screen, dpi);
 }
 #endif
 
+static void
+settings_update_color_scheme (GtkSettings *settings)
+{
+  gchar *dummy;
+
+  g_object_get (settings, "gtk-color-scheme", &dummy, NULL);
+
+  /* nothing to do here, the color hash is updated as a
+   * side effect of getting the color scheme
+   */
+
+  g_free (dummy);  
+}
+
+
+typedef struct {
+  GHashTable *color_hash;
+  GHashTable *tables[GTK_SETTINGS_SOURCE_APPLICATION + 1];
+  gchar *lastentry[GTK_SETTINGS_SOURCE_APPLICATION + 1];
+} ColorSchemeData;
+
+
+static gboolean
+add_color_to_hash (gchar      *name, 
+		   GdkColor   *color, 
+		   GHashTable *target)
+{
+  GdkColor *old, *new;
+
+  old = g_hash_table_lookup (target, name);
+  if (!old || !gdk_color_equal (old, color))
+    {
+      new = gdk_color_copy (color);
+      g_hash_table_insert (target, g_strdup (name), new);
+      
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
+static gboolean
+add_colors_to_hash_from_string (GHashTable  *hash, 
+				const gchar *colors)
+{
+  gchar *s, *p, *name;
+  GdkColor color;
+  gboolean changed = FALSE;
+  gchar *copy;
+
+  copy = g_strdup (colors);
+  s = copy;
+  while (s && *s)
+    {
+      name = s;
+      p = strchr (s, ':');
+      if (p)
+        {
+          *p = '\0';
+          p++;
+        }
+      else
+	break;
+
+      while (*p == ' ')
+        p++;
+
+      s = p;
+      while (*s) 
+	{
+	  if (*s == '\n' || *s == ';')
+	    {
+	      *s = '\0';
+	      s++;
+	      break;
+	    }
+	  s++;
+	}
+
+      if (gdk_color_parse (p, &color))
+	changed = add_color_to_hash (name, &color, hash);
+    }
+
+  g_free (copy);
+
+  return changed;
+}
+
+static gboolean
+update_color_hash (ColorSchemeData   *data,
+		   const gchar       *str, 
+		   GtkSettingsSource  source)
+{
+  gboolean changed = FALSE;
+  gint i;
+
+  if ((str == NULL || *str == '\0') && 
+      (data->lastentry[source] == NULL || data->lastentry[source][0] == '\0'))
+    return FALSE;
+
+  if (str && data->lastentry[source] && strcmp (str, data->lastentry[source]) == 0)
+    return FALSE;
+
+  /* For the RC_FILE source we merge the values rather than over-writing 
+   * them, since multiple rc files might define independent sets of colors
+   */
+  if ((source != GTK_SETTINGS_SOURCE_RC_FILE) && 
+      data->tables[source] && g_hash_table_size (data->tables[source]) > 0)
+    {
+      g_hash_table_unref (data->tables[source]);
+      data->tables[source] = NULL;
+      changed = TRUE; /* We can't rely on the code below since str might be "" */
+    }
+
+  if (data->tables[source] == NULL)
+    data->tables[source] = g_hash_table_new_full (g_str_hash, g_str_equal, 
+						  g_free,
+						  (GDestroyNotify) gdk_color_free);
+
+  g_free (data->lastentry[source]);
+  data->lastentry[source] = g_strdup (str);
+  
+  changed |= add_colors_to_hash_from_string (data->tables[source], str);
+
+  if (!changed)
+    return FALSE;
+    
+  /* Rebuild the merged hash table. */
+  g_hash_table_unref (data->color_hash);
+  data->color_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
+					    (GDestroyNotify) gdk_color_free);
+  for (i = 0; i <= GTK_SETTINGS_SOURCE_APPLICATION; i++)
+    {
+      if (data->tables[i])
+	g_hash_table_foreach (data->tables[i], (GHFunc) add_color_to_hash,
+			      data->color_hash);
+    }
+
+  return TRUE;
+}
+
+static void
+color_scheme_data_free (ColorSchemeData *data)
+{
+  gint i;
+
+  g_hash_table_unref (data->color_hash);
+
+  for (i = 0; i <= GTK_SETTINGS_SOURCE_APPLICATION; i++)
+    {
+      if (data->tables[i])
+	g_hash_table_unref (data->tables[i]);
+      g_free (data->lastentry[i]);
+    }
+
+  g_free (data);
+}
+
+static void
+merge_color_scheme (GtkSettings       *settings, 
+		    GValue            *value, 
+		    GtkSettingsSource  source)
+{
+  ColorSchemeData *data;
+  const gchar *colors;
+
+  colors = g_value_get_string (value);
+
+  data = (ColorSchemeData *) g_object_get_data (G_OBJECT (settings),
+						"gtk-color-scheme");
+  if (!data)
+    {
+      data = g_new0 (ColorSchemeData, 1);
+      data->color_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, 
+						(GDestroyNotify) gdk_color_free);
+      g_object_set_data_full (G_OBJECT (settings), "gtk-color-scheme",
+			      data, (GDestroyNotify) color_scheme_data_free);
+    }
+  
+  if (update_color_hash (data, colors, source))
+    g_object_notify (G_OBJECT (settings), "color-hash");
+}
+
+static GHashTable *
+get_color_hash (GtkSettings *settings)
+{
+  ColorSchemeData *data;
+
+  if (!g_object_get_data (G_OBJECT (settings), "gtk-color-scheme"))
+    settings_update_color_scheme (settings);
+  
+  data = (ColorSchemeData *)g_object_get_data (G_OBJECT (settings), 
+					       "gtk-color-scheme");
+
+  if (!data)
+    return NULL;
+
+  return data->color_hash;
+}
+
+static void 
+append_color_scheme (gpointer key,
+		     gpointer value,
+		     gpointer data)
+{
+  gchar *name = (gchar *)key;
+  GdkColor *color = (GdkColor *)value;
+  GString *string = (GString *)data;
+
+  g_string_append_printf (string, "%s: #%04x%04x%04x\n",
+			  name, color->red, color->green, color->blue);
+}
+
+static gchar *
+get_color_scheme (GtkSettings *settings)
+{
+  ColorSchemeData *data;
+  GString *string;
+  
+  data = (ColorSchemeData *) g_object_get_data (G_OBJECT (settings),
+						"gtk-color-scheme");
+
+  string = g_string_new ("");
+
+  g_hash_table_foreach (data->color_hash, append_color_scheme, string);
+
+  return g_string_free (string, FALSE);
+}
+
+
 #define __GTK_SETTINGS_C__
 #include "gtkaliasdef.c"
